@@ -1,9 +1,8 @@
-/* Copyright (C) 2000 MySQL AB
+/* Copyright (c) 2000-2003, 2007 MySQL AB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -16,22 +15,28 @@
 
 
 #include <my_global.h>
-#include "my_sys.h"
+#include <my_sys.h>
+#include <zlib.h>
 
 /*
-  Calculate a long checksum for a memoryblock. Used to verify pack_isam 
-   
+  Calculate a long checksum for a memoryblock.
+
   SYNOPSIS
-    checksum()
-      mem	Pointer to memory block
-      count	Count of bytes 
+    my_checksum()
+      crc       start value for crc
+      pos       pointer to memory block
+      length    length of the block
 */
 
-ulong checksum(const byte *mem, uint count)
+ha_checksum my_checksum(ha_checksum crc, const uchar *pos, size_t length)
 {
-  ulong crc;
-  for (crc= 0; count-- ; mem++)
-    crc= ((crc << 1) + *((uchar*) mem)) +
-      test(crc & ((ulong) 1L << (8*sizeof(ulong)-1)));
+#ifdef NOT_USED
+  const uchar *end=pos+length;
+  for ( ; pos != end ; pos++)
+    crc=((crc << 8) + *((uchar*) pos)) + (crc >> (8*sizeof(ha_checksum)-8));
   return crc;
+#else
+  return (ha_checksum)crc32((uint)crc, pos, (uint)length);
+#endif
 }
+
