@@ -1,9 +1,8 @@
-/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+/* Copyright (c) 2000, 2001, 2003, 2005-2007 MySQL AB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -37,3 +36,37 @@ void free_list(I_List <i_string> *list)
   while ((tmp= list->get()))
     delete tmp;
 }
+
+
+base_list::base_list(const base_list &rhs, MEM_ROOT *mem_root)
+{
+  if (rhs.elements)
+  {
+    /*
+      It's okay to allocate an array of nodes at once: we never
+      call a destructor for list_node objects anyway.
+    */
+    first= (list_node*) alloc_root(mem_root,
+                                   sizeof(list_node) * rhs.elements);
+    if (first)
+    {
+      elements= rhs.elements;
+      list_node *dst= first;
+      list_node *src= rhs.first;
+      for (; dst < first + elements - 1; dst++, src= src->next)
+      {
+        dst->info= src->info;
+        dst->next= dst + 1;
+      }
+      /* Copy the last node */
+      dst->info= src->info;
+      dst->next= &end_of_list;
+      /* Setup 'last' member */
+      last= &dst->next;
+      return;
+    }
+  }
+  elements= 0;
+  first= &end_of_list;
+  last= &first;
+}
